# -*- coding: utf-8 -*-

import datetime

from atelier.sphinxconf import configure

configure(globals())
from lino.sphinxcontrib import configure

configure(globals())

project = 'getlino docs'
html_title = "getlino"
copyright = '2019-{} Rumma & Ko Ltd'.format(datetime.date.today().year)
extensions += ['lino.sphinxcontrib.logo']

html_context.update({
    'display_gitlab': True,
    'gitlab_user': 'lino-framework',
    'gitlab_repo': 'getlino',
})

#
# import sys,os
#
# sys.path.insert(0, os.path.abspath(".."))
# import getlino
#
# extensions = []
# intersphinx_mapping = {}
#
# from rstgen.sphinxconf import configure
# configure(globals())
#
# extensions += ['sphinx.ext.autosummary']
# extensions += ['sphinx.ext.intersphinx']
#
# # intersphinx_mapping = {}
# from rstgen.sphinxconf import interproject
# interproject.configure(globals(), 'atelier')
# intersphinx_mapping['cg'] = ('https://community.lino-framework.org/', None)
# intersphinx_mapping['lf'] = ('https://www.lino-framework.org/', None)
# intersphinx_mapping['book'] = ('https://dev.lino-framework.org/', None)
#
# # General information about the project.
# project = "getlino"
# copyright = '2019-2021 Rumma & Ko Ltd'
# from getlino import SETUP_INFO
# release = SETUP_INFO['version']
# version = '.'.join(release.split('.')[:2])
#
# # The language for content autogenerated by Sphinx. Refer to documentation
# # for a list of supported languages.
# #language = None
#
# # There are two options for replacing |today|: either, you set today to some
# # non-false value, then it is used:
# #today = ''
# # Else, today_fmt is used as the format for a strftime call.
# #today_fmt = '%B %d, %Y'
#
# # List of documents that shouldn't be included in the build.
# #unused_docs = []
#
# # List of directories, relative to source directory, that shouldn't be searched
# # for source files.
# exclude_trees = [
#   'include',
#   ]
#
# # The reST default role (used for this markup: `text`) to use for all documents.
# #default_role = None
#
# # If true, '()' will be appended to :func: etc. cross-reference text.
# #add_function_parentheses = True
#
# # If true, the current module name will be prepended to all description
# # unit titles (such as .. function::).
# #add_module_names = True
#
# # If true, sectionauthor and moduleauthor directives will be shown in the
# # output. They are ignored by default.
# #show_authors = False
#
# # The name of the Pygments (syntax highlighting) style to use.
# pygments_style = 'sphinx'
#
#
# # Options for HTML output
# # -----------------------
#
# # The style sheet to use for HTML and HTML Help pages. A file of that name
# # must exist either in Sphinx' static/ path, or in one of the custom paths
# # given in html_static_path.
# # html_style = 'default.css'
#
# # The name for this set of Sphinx documents.  If None, it defaults to
# # "<project> v<release> documentation".
# html_title = SETUP_INFO['name'] # u"DjangoSite"
#
# # A shorter title for the navigation bar.  Default is the same as html_title.
# #html_short_title = None
#
# # The name of an image file (relative to this directory) to place at the top
# # of the sidebar.
# #~ html_logo = 'logo.jpg'
# #~ html_logo = 'lino-logo-2.png'
#
# # The name of an image file (within the static path) to use as favicon of the
# # docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# # pixels large.
# #~ html_favicon = 'favicon.ico'
#
# # Add any paths that contain custom static files (such as style sheets) here,
# # relative to this directory. They are copied after the builtin static files,
# # so a file named "default.css" will overwrite the builtin "default.css".
# # html_static_path = ['.static']
#
# # If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# # using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'
# #~ last_updated = True
#
# # If true, SmartyPants will be used to convert quotes and dashes to
# # typographically correct entities.
# #~ html_use_smartypants = True
#
# # Custom sidebar templates, maps document names to template names.
# html_sidebars = {
#    '**': ['globaltoc.html', 'searchbox.html', 'links.html'],
# }
#
# # Additional templates that should be rendered to pages, maps page names to
# # template names.
# #~ html_additional_pages = {
#     #~ '*': 'links.html',
# #~ }
#
#
# # If false, no module index is generated.
# html_use_modindex = True
#
# # If false, no index is generated.
# html_use_index = True
#
# # If true, the index is split into individual pages for each letter.
# #html_split_index = False
#
# # If true, the reST sources are included in the HTML build as _sources/<name>.
# html_copy_source = False
#
# # If true, an OpenSearch description file will be output, and all pages will
# # contain a <link> tag referring to it.  The value of this option must be the
# # base URL from which the finished HTML is served.
# #html_use_opensearch = ''
# html_use_opensearch = 'http://getlino.lino-framework.org'
#
# # If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
# #html_file_suffix = ''
#
# # Output file base name for HTML help builder.
# htmlhelp_basename = 'getlino'
#
# extlinks.update(
#   # srcref=(getlino.srcref_url, ''),
#   djangoticket=('http://code.djangoproject.com/ticket/%s', 'Django ticket #'),
# )
#
# autosummary_generate = True
#
# todo_include_todos = True
#
# gettext_compact = True
#
# # extlinks.update(ticket=('https://jane.mylino.net/#/api/tickets/AllTickets/%s', '#'))
#
# suppress_warnings = ['image.nonlocal_uri']



"""
The icons downloaded by this script are Copyright (c) 2018 PrimeTek
See https://github.com/primefaces/primeicons/blob/master/LICENSE

The master copy of this file is in the book repository. Don't edit its copy in a
slave repository.  See https://dev.lino-framework.org/writedocs/shared.html

To see a list of available primefaces icons, visit:
https://github.com/primefaces/primeicons/tree/master/raw-svg


"""
import requests
from pathlib import Path

# TODO: from lino.core.constants import ICON_NAMES
ICON_NAMES = ("external-link", "filter", "plus", "plus-circle", "user",
              "refresh", "trash", "bell", "times-circle", "envelope", "copy",
              "save", "home", "bars", "eject", "sliders-h")

dest = Path(__file__).parent.absolute()
root = "https://raw.githubusercontent.com/primefaces/primeicons/master/raw-svg"
found = []
for name in ICON_NAMES:
    filename = name + ".svg"
    target = dest / filename
    if target.exists():
        found.append(filename)
    else:
        url = "{}/{}".format(root, filename)
        print("Downloading {} from primeicons master...".format(filename))
        r = requests.get(url, stream=True)
        with open(target, "wb") as f:
            for chunk in r.iter_content(chunk_size=16 * 1024):
                f.write(chunk)
if len(found):
    print("No need to download {}".format(", ".join(found)))



"""

A collection of scripts used to install Lino on a computer system.


.. autosummary::
   :toctree:

   configure
   startsite
   utils
   cli

"""

__version__ = '25.7.0'

intersphinx_urls = dict(docs="https://lino-framework.gitlab.io/getlino")
srcref_url = 'https://gitlab.com/lino-framework/getlino/blob/master/%s'
doc_trees = ['docs']



# Copyright 2019-2025 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

import os
from os.path import join
import distro
import click
from pathlib import Path

from synodal import KNOWN_REPOS, FRONT_ENDS

from .utils import CONFIG, CONF_FILES, DEFAULTSECTION
from .utils import DB_ENGINES, BATCH_HELP
from .utils import Installer, ifroot, default_db_engine, resolve_db_engine
from .utils import which_certbot
from .utils import WEB_SERVERS, resolve_web_server, GITLAB_CI

HEALTHCHECK_NAME = '/usr/local/bin/healthcheck.sh'
RELOAD_NAME = '/usr/local/bin/reload_services.sh'

CERTBOT_AUTO_RENEW = """\
# generated by getlino
0 0,12 * * * root python -c 'import random; import time; time.sleep(random.random() * 3600)' && {} renew
"""

MONIT_CONF = """\
# generated by getlino
set alert root@localhost with reminder on 2 cycles
check program status with path {}
    if status != 0 then alert
check device ROOT  with path /
    if SPACE usage > 95% then alert
check system $HOST
    if memory usage > 75% for 5 cycles then alert
""".format(HEALTHCHECK_NAME)

# HEALTHCHECK_CRON_JOB = """\
# #!/bin/bash
# /usr/local/bin/healthcheck.sh > /dev/null
# # we redirect STDOUT to null, but not STDERR because if there is an error we want cron to notify the admins.
# """

LIBREOFFICE_SUPERVISOR_CONF = """\
# generated by getlino
[program:libreoffice]
command = libreoffice --accept="socket,host=127.0.0.1,port=8100;urp;" --nologo --headless --nofirststartwizard
umask = 0002
"""

SHARED_SETTINGS = r"""\
# generated by getlino
# this is the shared settings file, imported by every Lino site on this server
ADMINS = [
  ["{admin_name}", "{admin_email}"]
]
EMAIL_HOST = 'localhost'
SERVER_EMAIL = 'noreply@{server_domain}'
DEFAULT_FROM_EMAIL = 'noreply@{server_domain}'
STATIC_ROOT = 'env/static_root'
TIME_ZONE = "{time_zone}"
# USE_TZ = True
# TIME_ZONE = 'UTC'
# TIME_ZONE = 'Europe/Brussels'
# TIME_ZONE = 'Europe/Tallinn'
# List of accepted time zones:
# import pytz; print("\n".join(pytz.all_timezones))
"""

BASH_ALIASES = """\
#!/bin/bash
# generated by getlino
alias a='. env/bin/activate'
alias ll='ls -alF'
alias pm='python manage.py'
alias runserver='LINO_LOGLEVEL=DEBUG python manage.py runserver'
function pywhich() {{
  python -c "import $1; print($1.__file__)"
}}
"""
BASH_ALIASES_GO = """
function go() {{
    for BASE in {go_bases}
    do
      if [ -d $BASE/$1 ] ; then
        cd $BASE/$1;
        return;
      fi
    done
    echo Oops: no project $1
    return -1
}}
"""

BASH_ALIASES_DEV = """
alias pp='per_project'
alias ci='inv ci'
alias dt='doctest'
"""

# The configure command will be decorated below. We cannot use decorators
# because we define the list of options in CONFIGURE_OPTIONS because we need
# that list also for asking questions using the help text.

CONFIGURE_OPTIONS = []


def add(spec, default, help, type=None, root_only=False):
    """
    :param root_only: if user is not root, do not ask for the user for the choice.

    """
    kwargs = dict()
    kwargs.update(help=help)
    if type is not None:
        kwargs.update(type=type)
    o = click.Option([spec], **kwargs)
    o.root_only = root_only
    o.default = DEFAULTSECTION.get(o.name, default)  # ~/.getlino.conf
    CONFIGURE_OPTIONS.append(o)


def default_sites_base():
    return ifroot('/usr/local/lino', os.path.expanduser('~/lino'))


def default_shared_env():
    return ifroot('', os.environ.get('VIRTUAL_ENV', ''))
    # return os.environ.get('VIRTUAL_ENV', '/usr/local/lino/venvs')


# def default_repos_base():
#     #if default_shared_env():
#     return ifroot('/usr/local/lino/repositories', os.path.expanduser('~/lino/repositories'))
#     #return ''

# must be same order as in signature of configure command below
# add('--prod/--no-prod', True, "Whether this is a production server")
add('--sites-base', default_sites_base,
    'Base directory for Lino sites on this server')
add('--local-prefix', 'lino_local',
    "Prefix for local server-wide importable packages")
add('--shared-env', default_shared_env,
    "Root directory of your shared virtualenv")
add('--repos-base', '', "Base directory for shared code repositories")
# add('--repos-base', default_repos_base, "Base directory for shared code repositories")
add('--clone/--no-clone', False,
    "Clone all contributor repositories and install them to the shared-env")
# add('--branch', 'master', "The git branch to use for --clone")
add('--webdav/--no-webdav',
    True,
    "Whether to enable webdav on new sites",
    root_only=True)
add('--backups-base',
    '/var/backups/lino',
    'Base directory for backups',
    root_only=True)
# add('--log-base',
#     '/var/log/lino',
#     'Base directory for log files',
#     root_only=True)
add('--usergroup',
    'www-data',
    "User group for files to be shared with the web server",
    root_only=True)
add('--supervisor-dir',
    '/etc/supervisor/conf.d',
    "Directory for supervisor config files",
    root_only=True)
add('--env-link', 'env', "link to virtualenv (relative to project dir)")
add('--repos-link', 'repositories',
    "link to code repositories (relative to virtualenv)")
add('--appy/--no-appy',
    ifroot,
    "Whether this server provides appypod and LibreOffice",
    root_only=True)
add('--journal/--no-journal',
    ifroot,
    "Whether Lino sites should log to systemd journal",
    root_only=True)
add('--weasyprint/--no-weasyprint',
    ifroot,
    "Whether this server provides weasyprint",
    root_only=True)
add('--redis/--no-redis', ifroot, "Whether this server provides redis")
add('--devtools/--no-devtools', lambda: not ifroot(),
    "Whether to install development tools (build docs and run tests)")
add('--server-domain', 'localhost', "Domain name of this server")
add('--https/--no-https',
    False,
    "Whether this server uses secure http",
    root_only=True)
add('--ldap/--no-ldap',
    False,
    "Whether this server works as an LDAP server",
    root_only=True)
add('--monit/--no-monit',
    True,
    "Whether this server uses monit",
    root_only=True)
add('--web-server',
    '',
    "Which web server to use here.",
    click.Choice([e.name for e in WEB_SERVERS] + ['']),
    root_only=True)
add('--db-engine', default_db_engine, "Default database engine for new sites.",
    click.Choice([e.name for e in DB_ENGINES]))
add('--db-port', '', "Default database port to use for new sites.")
add('--db-host', 'localhost', "Default database host name for new sites.")
add(
    '--db-user', '',
    "Default database user name for new sites. Leave empty to use the project name."
)
add(
    '--db-password', '',
    "Default database password for new sites. Leave empty to generate a secure password."
)
add('--admin-name', 'Joe Dow', "The full name of the server administrator")
add('--admin-email', 'joe@example.com',
    "The email address of the server administrator")
add('--time-zone', 'Europe/Brussels', "The TIME_ZONE to set on new sites")
add('--linod/--no-linod', True, "Whether new sites use linod", root_only=True)
add('--languages', 'en', "The languages to set on new sites")
add('--front-end', 'lino_react.react', "The front end to use on new sites",
    click.Choice(FRONT_ENDS.keys()))


def configure(ctx, batch, sites_base, local_prefix, shared_env, repos_base,
              # clone, webdav, backups_base, log_base, usergroup, supervisor_dir,
              clone, webdav, backups_base, usergroup, supervisor_dir,
              env_link, repos_link, appy, journal, weasyprint, redis, devtools, server_domain,
              https, ldap, monit, db_engine, db_port, db_host, db_user,
              db_password, admin_name, admin_email, time_zone, linod,
              languages, front_end, web_server):
    """
    Configure this machine to become a Lino production server.
    """

    # if len(FOUND_CONFIG_FILES) > 1:
    #     # reconfigure is not yet supported
    #     raise click.UsageError("Found multiple config files: {}".format(
    #         FOUND_CONFIG_FILES))

    i = Installer(batch)
    context = {}
    context.update(locals())
    context.update({
        "prjname": '',
        "appname": '',
        "project_dir": '',
        "repo_nickname": '',
        "app_package": '',
        "app_settings_module": '',
        "django_settings_module": '',
        "dev_packages": ' '.join([a.nickname for a in KNOWN_REPOS if a.git_repo]),
        "pip_packages": '',
        "python_path": ''
    })

    conffile = ifroot(CONF_FILES[0], CONF_FILES[1])
    click.echo("This will write to configuration file {}".format(conffile))

    # before asking questions check whether we will be able to store them
    pth = os.path.dirname(conffile)
    if not os.path.exists(pth):
        os.makedirs(pth, exist_ok=True)
    if not os.access(pth, os.W_OK):
        raise click.ClickException(
            "No write permission for directory {}".format(pth))
    if os.path.exists(conffile) and not os.access(conffile, os.W_OK):
        raise click.ClickException(
            "No write permission for file {}".format(conffile))

    if clone and not shared_env:
        shared_env = os.environ.get('VIRTUAL_ENV', '')

    # if shared_env: not sure whether this is a good idea
    #     shared_env = os.path.abspath(shared_env)

    i.ask_options(CONFIGURE_OPTIONS, locals())

    # update context and local vars after interactive prompt
    context.update(DEFAULTSECTION)
    clone = DEFAULTSECTION.getboolean('clone')
    shared_env = DEFAULTSECTION.get('shared_env')
    db_user = DEFAULTSECTION.get('db_user')
    db_password = DEFAULTSECTION.get('db_password')
    db_engine = resolve_db_engine(DEFAULTSECTION.get('db_engine'))
    web_server = resolve_web_server(DEFAULTSECTION.get('web_server'))
    appy = DEFAULTSECTION.getboolean('appy')
    weasyprint = DEFAULTSECTION.getboolean('weasyprint')
    supervisor = web_server or appy
    devtools = DEFAULTSECTION.getboolean('devtools')

    # combined validations
    if clone and not shared_env:
        raise click.ClickException("Cannot use --clone without --shared-env")
    if db_user and not db_password:
        raise click.Error(
            "If you set a shared --db-user you must also set a shared --db-password"
        )

    if not i.yes_or_no("Write above options to {} ?".format(conffile)):
        raise click.Abort()

    with open(conffile, 'w') as fd:
        CONFIG.write(fd)
    click.echo("Wrote config file " + conffile)

    if not i.yes_or_no("Start installing Lino according to above settings?"):
        raise click.Abort()

    i.apt_install("git python3 python3-dev python3-setuptools python3-pip")
    i.apt_install("libffi-dev libssl-dev")  # maybe needed for weasyprint
    i.apt_install("build-essential")  # needed for installing mysqlclient
    i.apt_install("swig")  # required to install eidreader
    i.apt_install(
        "libldap2-dev libsasl2-dev")  # Needed e.g. by django_auth_ldap
    # i.apt_install("python2.7-dev libldap2-dev libsasl2-dev ldap-utils lcov") # Needed by Ldap package
    # i.runcmd("sudo DEBIAN_FRONTEND=noninteractive apt-get -yq install slapd")

    if supervisor:
        i.apt_install("supervisor")

    if web_server and ifroot():
        i.apt_install("cron")
        i.apt_install(web_server.apt_packages)
        i.apt_install("logrotate")
        i.must_restart(web_server.service)

    if devtools:
        i.apt_install("subversion graphviz sqlite3")

    if DEFAULTSECTION.getboolean('monit'):
        i.apt_install("monit")

    if DEFAULTSECTION.getboolean('redis') and not GITLAB_CI:
        i.apt_install("redis-server")
        i.must_restart('redis-server')

    if not GITLAB_CI:
        i.apt_install(db_engine.apt_packages)

    if web_server:
        i.apt_install(web_server.apt_packages)

    if db_engine.service and not GITLAB_CI:
        i.must_restart(db_engine.service)

    if appy:
        i.apt_install("libreoffice-common python3-uno")
        i.apt_install("tidy")

    if journal:
        i.apt_install("libsystemd-dev")

    if weasyprint:
        # https://doc.courtbouillon.org/weasyprint/stable/first_steps.html
        i.apt_install(
            "python3-pip libpango-1.0-0 libpangoft2-1.0-0 libharfbuzz-subset0")

    if DEFAULTSECTION.getboolean('ldap'):
        i.apt_install("slapd ldap-utils")

    if ifroot():
        # for k in ("log_base", "backups_base"):
        for k in ("backups_base",):
            pth = DEFAULTSECTION.get(k)
            if not pth:
                print("Strange: {} is empty...".format(k))
                continue
            if not os.path.exists(pth):
                if batch or i.yes_or_no("Create {} {} ?".format(k, pth),
                                        default=True):
                    os.makedirs(pth, exist_ok=True)
            i.check_permissions(pth)
        i.apt_install("zip")

    if ifroot() and DEFAULTSECTION.getboolean('monit'):
        # click.echo("20200727 Install monit...")
        # Debian buster didn't include monit for administrative reasons,
        # so we must add a backport.
        if distro.id() == "debian" and distro.codename() == "buster":
            if batch or i.yes_or_no("Add backport for monit on Debian buster?",
                                    default=True):
                click.echo("Adding backport for monit on Debian buster...")
                with i.override_batch(True):
                    pth = Path("/etc/apt/sources.list.d/buster-backports.list")
                    myline = "deb http://ftp.de.debian.org/debian buster-backports main"
                    if pth.is_file():
                        content = pth.open().read()
                        if not myline in content:
                            content += "\n" + myline + "\n"
                    else:
                        content = myline + "\n"
                    i.write_file(pth, content)
                    # i.runcmd('printf "%s\\n" "deb http://ftp.de.debian.org/debian buster-backports main" | sudo tee /etc/apt/sources.list.d/buster-backports.list')
                    # i.runcmd('echo "deb http://ftp.de.debian.org/debian buster-backports main" >> /etc/apt/sources.list.d/buster-backports.list')

    i.run_apt_install()
    i.restart_services()

    if DEFAULTSECTION.get('db_user') and not GITLAB_CI:
        db_engine.setup_user(i, context)

    go_bases = []

    if clone:
        click.echo("Installing repositories for shared-env...")
        if not shared_env:
            raise click.ClickException(
                "With --clone you must either "
                "have a virtualenv activated or specify one with --shared-env")

        repos_base = DEFAULTSECTION.get('repos_base')
        if not repos_base:
            repos_base = join(shared_env, DEFAULTSECTION.get('repos_link'))

        if not os.path.exists(repos_base):
            if batch or i.yes_or_no(
                    "Create base directory for repositories {} ?".format(
                        repos_base),
                    default=True):
                os.makedirs(repos_base, exist_ok=True)
        i.check_permissions(repos_base)

        i.check_virtualenv(shared_env, context)
        os.chdir(repos_base)
        repos = [r for r in KNOWN_REPOS if r.git_repo]
        if batch or i.yes_or_no(
                "Clone repositories to {} ?".format(repos_base), default=True):
            with i.override_batch(True):
                for repo in repos:
                    i.clone_repo(repo)
        if batch or i.yes_or_no(
                "Install cloned repositories to {} ?".format(shared_env),
                default=True):
            # getlino must not reinstall itself because that causes at least
            # TemplateNotFound issues
            with i.override_batch(True):
                for repo in repos:
                    if repo.package_name:
                        if repo.package_name != "getlino":
                            i.install_repo(repo, shared_env)
        go_bases.append(repos_base)
        for repo in repos:
            if repo.package_name:
                name = repo.package_name.replace("-", "_")
                pth = Path(repos_base, repo.nickname, name, "projects")
                if pth.exists():
                    go_bases.append(str(pth))
                # else:
                #     print("20240708 no projects dir in", repo)

    pth = DEFAULTSECTION.get('sites_base')
    if not os.path.exists(pth):
        if batch or i.yes_or_no(
                "Create base directory for sites {} ?".format(pth),
                default=True):
            os.makedirs(pth, exist_ok=True)
    i.check_permissions(pth)

    local_prefix = DEFAULTSECTION.get('local_prefix')
    pth = join(DEFAULTSECTION.get('sites_base'), local_prefix)
    if os.path.exists(pth):
        i.check_permissions(pth)
    elif batch or i.yes_or_no(
            "Create shared settings package {} ?".format(pth), default=True):
        os.makedirs(pth, exist_ok=True)
        i.check_permissions(pth)
    with i.override_batch(True):
        i.check_permissions(pth)
        fn = join(pth, '__init__.py')
        i.write_file(fn, '')
        i.check_permissions(fn)
    fn = join(pth, 'settings.py')
    i.write_file(fn, SHARED_SETTINGS.format(**DEFAULTSECTION))
    i.check_permissions(fn)
    go_bases.append(pth)

    pth = ifroot('/etc/getlino/lino_bash_aliases',
                 os.path.expanduser('~/.lino_bash_aliases'))
    ctx = dict(DEFAULTSECTION)
    content = BASH_ALIASES.format(**ctx)
    if len(go_bases):
        ctx.update(go_bases=" ".join(go_bases))
        content += BASH_ALIASES_GO.format(**ctx)
    if devtools:
        content += BASH_ALIASES_DEV
    i.write_file(pth, content)
    i.check_permissions(pth, writeable=False)
    click.echo(
        "Note: please add manually the following line to your .bashrc file:\nsource {}"
        .format(pth))

    if ifroot():
        if supervisor:
            i.write_logrotate_conf('supervisor.conf',
                                   '/var/log/supervisor/supervisord.log')
            i.must_restart('supervisor')

        i.jinja_write(RELOAD_NAME, **context)
        i.check_permissions(RELOAD_NAME, executable=True)

        if DEFAULTSECTION.getboolean('monit'):
            i.jinja_write(HEALTHCHECK_NAME, **context)
            i.check_permissions(HEALTHCHECK_NAME, executable=True)
            i.write_file('/etc/monit/conf.d/lino.conf', MONIT_CONF)
            # seems that monit creates its own logrotate config file
            # i.write_logrotate_conf(
            #     'monit.conf', '/var/log/monit.log')

        # i.write_cron_job('healthcheck',
        #                  HEALTHCHECK_CRON_JOB.format(**context),
        #                  freq="hourly")

        if supervisor and appy:
            i.write_supervisor_conf(
                'libreoffice.conf',
                LIBREOFFICE_SUPERVISOR_CONF.format(**DEFAULTSECTION))
            i.must_restart('supervisor')

        # if DEFAULTSECTION.get('db_engine') == 'mysql':
        #     i.runcmd("mysql_secure_installation")
        # not tested because it interactively asks for root password

        if DEFAULTSECTION.getboolean('https'):
            certbot_cmd = which_certbot()
            if certbot_cmd:
                click.echo("{} already installed".format(certbot_cmd))
            if certbot_cmd is None:
                if batch or i.yes_or_no("Install certbot?", default=True):
                    # i.apt_install("certbot python-certbot-nginx")
                    i.apt_install("snapd")
                    i.run_apt_install()
                    i.runcmd_sudo("snap install core")
                    i.runcmd_sudo("snap refresh core")
                    i.runcmd_sudo("snap install --classic certbot")
                    i.runcmd_sudo("ln -s /snap/bin/certbot /usr/bin/certbot")
                    certbot_cmd = "certbot"
            # if certbot_cmd is None:
            #     if batch or i.yes_or_no("Install certbot-auto?", default=True):
            #         with i.override_batch(True):
            #             i.runcmd("wget https://dl.eff.org/certbot-auto")
            #             i.runcmd("mv certbot-auto /usr/local/bin/certbot-auto")
            #             i.runcmd("chown root /usr/local/bin/certbot-auto")
            #             i.runcmd("chmod 0755 /usr/local/bin/certbot-auto")
            #             i.runcmd("certbot-auto -n")
            #             i.runcmd("certbot-auto register --agree-tos -m {} -n".format(DEFAULTSECTION.get('admin_email')))
            #         certbot_cmd = "/usr/local/bin/certbot-auto"
            if certbot_cmd:
                if batch or i.yes_or_no(
                        "Set up automatic certificate renewal?", default=True):
                    i.write_file('/etc/cron.d/getlino-certbot.conf',
                                 CERTBOT_AUTO_RENEW.format(certbot_cmd))

        if DEFAULTSECTION.getboolean('ldap'):
            i.runcmd("dpkg-reconfigure slapd")

    i.restart_services()

    click.echo("getlino configure completed.")


params = [
    click.Option(['--batch/--no-batch'], default=False, help=BATCH_HELP),
] + CONFIGURE_OPTIONS
configure = click.pass_context(configure)
configure = click.Command('configure',
                          callback=configure,
                          params=params,
                          help=configure.__doc__)



# Copyright 2019-2022 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from rstgen.confparser import ConfigParser
from configparser import NoOptionError, NoSectionError
from .utils import GITLAB_CI
from .utils import JINJA_ENV
from .utils import which_certbot, resolve_web_server
from .utils import default_db_engine, resolve_db_engine
from .utils import Installer, ifroot, has_usergroup
from .utils import DB_ENGINES, BATCH_HELP
from .utils import DEFAULTSECTION
from synodal import REPOS_DICT, KNOWN_REPOS, FRONT_ENDS
from os.path import join
import os
import sys
import shutil
import secrets
import click
from pathlib import Path
from logging import getLogger
from urllib.parse import urlparse

logger = getLogger(__file__)


APPNAMES = [a.nickname for a in KNOWN_REPOS if a.settings_module]

UWSGI_SUPERVISOR_CONF = """\
# generated by getlino
[program:{prjname}-uwsgi]
command = /usr/bin/uwsgi --ini {project_dir}/nginx/uwsgi.ini --ignore-sigpipe
user = {usergroup}
umask = 0002
stopsignal = QUIT
"""

LINOD_SUPERVISOR_CONF = """\
# generated by getlino
[program:linod-{prjname}]
command={project_dir}/linod.sh
user = {usergroup}
umask = 0002
"""
LINOD_SH = """\
#!/bin/bash
set -e  # exit on error
PRJ={project_dir}
. $PRJ/{env_link}/bin/activate
exec python $PRJ/manage.py linod
"""

MAKE_SNAPSHOT_CRON_SH = """\
#!/bin/sh
# generated by getlino
sudo service supervisor stop
{project_dir}/make_snapshot.sh > /dev/null
sudo service supervisor start
"""

ONLY_ROOT = "Only root can do startsite with {}"


def default_shared_env():
    return DEFAULTSECTION.get('shared_env')


def default_site_domain():
    return DEFAULTSECTION.get('server_domain')

# The configure command will be decorated below. We cannot use decorators
# because we define the list of options in CONFIGURE_OPTIONS because we need
# that list also for asking questions using the help text.


STARTSITE_OPTIONS = []


def add(spec, default=None, help=None, type=None, required=None, root_only=False):
    """
    :param root_only: if user is not root, do not ask for the user for the choice.

    """
    kwargs = dict()
    kwargs.update(help=help)
    if type is not None:
        kwargs.update(type=type)
    if required is not None:
        kwargs.update(required=required)
    o = click.Option([spec], **kwargs)
    o.root_only = root_only
    o.default = DEFAULTSECTION.get(o.name, default)  # ~/.getlino.conf
    STARTSITE_OPTIONS.append(o)


add('--dev-repos',
    default='',
    help="List of packages for which to install development version")
add('--shared-env',
    default=default_shared_env,
    help="Directory with shared virtualenv")
add('--site-domain',  # site_domain
    default=default_site_domain,
    help="FQDN for this site (without 'http(s)://' prefix).")
add('--db-engine',
    help="Database engine to use.",
    type=click.Choice([e.name for e in DB_ENGINES]))
add('--db-port', help="Database port to use.")
add('--db-host',
    default='localhost',
    help="Database host name to use.")
add(
    '--db-user',
    help="Database user name to use. Leave empty to use the project name.")
add(
    '--db-password',
    help="Password for database user. Leave empty to generate a secure password.")


def startsite(
    ctx,
    appname,
    prjname,
    ini_file,
    batch,
    dev_repos,
    shared_env,
    db_engine,
    db_port,
    db_host,
    db_user,
    db_password,
    site_domain
):
    """
    Create a new Lino site.

    Two mandatory arguments must be given unless `--ini-file` is specified:

    APPNAME : The application to run on the new site. Say `getlino list` to see
    a list of choices.

    PRJNAME : The nickname for the new site. We recommend lower-case only and
    digits but no "-" or "_". Examples:  foo, foo2, mysite, first, The name
    must be unique for this Lino server and  will become a subdirectory of the
    `--sites-base` directory specified by `getlino configure`.


    """  # .format(appnames=' '.join(APPNAMES))
    if (not appname or not prjname) and not ini_file:
        click.echo(ctx.get_help(), color=ctx.color)
        ctx.exit()

    i = Installer(batch)

    if ini_file:
        assert appname is None and prjname is None
        main_cfg = ConfigParser()
        main_cfg.read_file(ini_file)
        try:
            appname = main_cfg.get('getlino', 'appname')
            app_package = prjname = main_cfg.get('getlino', 'prjname')
            git_repo = main_cfg.get('getlino', 'git_repo')
            repo_nickname = git_repo.split('.git')[0].split('/')[-1]
            settings_module = main_cfg.get('getlino', 'settings_module')
            # try:
            #     git_branch = main_cfg.get('getlino', 'git_branch')
            # except NoOptionError as e:
            #     git_branch = DEFAULTSECTION.get('branch')
            # try:
            #     frontend_branch = main_cfg.get('getlino', 'frontend_branch')
            # except NoOptionError as e:
            #     frontend_branch = DEFAULTSECTION.get('branch')
            try:
                front_end = main_cfg.get('getlino', 'front_end')
            except NoOptionError as e:
                front_end = None
            try:
                use_django_settings = main_cfg.get('getlino',
                                                   'use_django_settings')
            except NoOptionError as e:
                use_django_settings = False
        except (NoSectionError, NoOptionError) as e:
            with open(Path(__file__).parent / 'templates' / 'lino.ini',
                      'r') as f:
                click.echo((
                    "INI must contain a 'getlino' section (with optional 'front_end').\nSee DEFAULT ini format: \n\n"
                    + f.read()),
                           color=ctx.color)
            ctx.exit()

        if not front_end and not use_django_settings:
            front_end = REPOS_DICT.get(DEFAULTSECTION.get('front_end'), None)

    i.ask_options(STARTSITE_OPTIONS, locals())

    # update context and local vars after interactive prompt
    sites_base = DEFAULTSECTION.get('sites_base')
    local_prefix = DEFAULTSECTION.get('local_prefix')
    python_path_root = join(sites_base, local_prefix)
    project_dir = join(python_path_root, prjname)
    if not os.path.exists(project_dir):
        os.makedirs(project_dir, exist_ok=True)
    i.check_permissions(project_dir)
    if shared_env:
        envdir = shared_env
    else:
        envdir = join(project_dir, DEFAULTSECTION.get('env_link'))

    def get_dev_repos_dir():
        full_repos_dir = DEFAULTSECTION.get('repos_base')
        if not full_repos_dir:
            full_repos_dir = join(envdir, DEFAULTSECTION.get('repos_link'))
            if not os.path.exists(full_repos_dir):
                os.makedirs(full_repos_dir, exist_ok=True)
        i.check_permissions(full_repos_dir)
        return full_repos_dir

    def change_directory_into_dev_repos():
        os.chdir(get_dev_repos_dir())

    if ini_file:
        # def clone_pull_install_repo(repo_nn, repo_remote, branch, settings_module=None):
        def clone_pull_install_repo(repo_nn,
                                    repo_remote,
                                    settings_module=None):
            clonable = True
            if os.path.exists(join(project_dir, repo_nn)):
                os.chdir(join(project_dir, repo_nn))
                i.runcmd("git pull --rebase -q")
                clonable = False

            if clonable and os.path.exists(join(get_dev_repos_dir(), repo_nn)):
                os.chdir(join(get_dev_repos_dir(), repo_nn))
                i.runcmd("git pull --rebase -q")
                clonable = False

            if clonable:
                os.chdir(project_dir)
                # i.runcmd("git clone -q --depth 1 -b {} {} {}".format(branch, repo_remote, repo_nn))
                i.runcmd("git clone -q {} {}".format(repo_remote, repo_nn))
                for f in ['setup.py', 'pyproject.toml']:
                    if os.path.exists(join(project_dir, repo_nn, f)):
                        shutil.move(join(project_dir, repo_nn),
                                    join(get_dev_repos_dir(), repo_nn))
                        change_directory_into_dev_repos()
                        i.run_in_env(envdir,
                                     "pip install -q -e {}".format(repo_nn))
                        break

            if os.path.exists(join(project_dir, repo_nn)):
                init_file = join(project_dir, repo_nn, '__init__.py')
                if not os.path.exists(init_file):
                    with open(init_file, 'w') as f:
                        f.write("# Created by getlino\n")
                if settings_module is not None:
                    return "{}.{}.{}.{}".format(local_prefix, prjname,
                                                repo_nickname, settings_module)

        # sm = clone_pull_install_repo(repo_nickname, git_repo, git_branch, settings_module)
        sm = clone_pull_install_repo(repo_nickname, git_repo, settings_module)
        if sm is not None:
            settings_module = sm
        if front_end and isinstance(front_end, str):
            u = urlparse(front_end)
            if all((u.scheme, u.netloc)):
                # clone_pull_install_repo(front_end.split('.git')[0].split('/')[-1], front_end, frontend_branch)
                clone_pull_install_repo(
                    front_end.split('.git')[0].split('/')[-1], front_end)

    # if len(FOUND_CONFIG_FILES) == 0:
    #     raise click.UsageError(
    #         "This server is not yet configured. Did you run `sudo -H getlino configure`?")

    # if os.path.exists(prjpath):
    #     raise click.UsageError("Project directory {} already exists.".format(prjpath))

    web_server = resolve_web_server(DEFAULTSECTION.get('web_server'))
    if web_server and not ifroot():
        raise click.ClickException(ONLY_ROOT.format(web_server))

    https = DEFAULTSECTION.getboolean('https')
    linod = DEFAULTSECTION.getboolean('linod')
    journal = DEFAULTSECTION.getboolean('journal')
    if linod and web_server and ifroot() and web_server.name == 'nginx':
        asgi_server = True
    else:
        asgi_server = False

    # prod = DEFAULTSECTION.getboolean('prod')
    # contrib = DEFAULTSECTION.getboolean('contrib')
    # shared_env = DEFAULTSECTION.get('shared_env')
    admin_name = DEFAULTSECTION.get('admin_name')
    admin_email = DEFAULTSECTION.get('admin_email')
    server_domain = DEFAULTSECTION.get('server_domain')
    # if web_server:
    #     server_domain = prjname + "." + server_domain
    server_url = ("https://" if https else "http://") + site_domain
    secret_key = secrets.token_urlsafe(20)

    db_engine = resolve_db_engine(db_engine or DEFAULTSECTION.get('db_engine')
                                  or default_db_engine())

    if db_engine.needs_root and not ifroot():
        raise click.ClickException(ONLY_ROOT.format(db_engine))

    db_host = db_host or DEFAULTSECTION.get('db_host')
    db_port = db_port or DEFAULTSECTION.get(
        'db_port') or db_engine.default_port

    usergroup = DEFAULTSECTION.get('usergroup')

    if not ini_file:
        app = REPOS_DICT.get(appname, None)
        if app is None:
            raise click.ClickException(
                "Invalid application nickname '{}'".format(appname))
        settings_module = app.settings_module
        if not settings_module:
            raise click.ClickException(
                "{} is a library, not an application".format(appname))

        front_end = FRONT_ENDS.get(DEFAULTSECTION.get('front_end'), None)
        if front_end is None:
            raise click.ClickException(
                "Invalid front_end name '{}''".format(front_end))

        if dev_repos:
            for k in dev_repos.split():
                repo = REPOS_DICT.get(k, None)
                if repo is None or not repo.git_repo:
                    nicknames = ' '.join(
                        [r.nickname for r in KNOWN_REPOS if r.git_repo])
                    raise click.ClickException(
                        "Invalid repository name {}. "
                        "Allowed names are one or more of ({})".format(
                            k, nicknames))

        # if not i.check_overwrite(project_dir):
        #     raise click.Abort()

        # if not i.asroot and not shared_env:
        #     raise click.ClickException(
        #         "Cannot startsite in a development environment without a shared-env!")

        app_package = app.package_name
        # app_package = app.settings_module.split('.')[0]
        repo_nickname = app.git_repo.split('/')[-1]

    context = {}
    context.update(DEFAULTSECTION)
    pip_packages = set()
    if not ini_file:  # and not shared_env:
        if app.nickname not in dev_repos:
            pip_packages.add(app.package_name)
        if front_end.nickname not in dev_repos:
            pip_packages.add(front_end.package_name)

        # 20221223 pull updates internal dependencies automatically
        for nickname in app.extra_deps:
            if nickname not in dev_repos:
                pip_packages.add(REPOS_DICT[nickname].package_name)

    for pkgname in db_engine.python_packages.split():
        pip_packages.add(pkgname)

    if journal:
        pip_packages.add('systemd-python')

    if linod and asgi_server:
        pip_packages.add('daphne')
        i.apt_install('redis-server')

    context.update({
        "prjname": prjname,
        "appname": appname,
        "project_dir": project_dir,
        "repo_nickname": repo_nickname,
        "app_package": app_package,
        "app_settings_module": settings_module,
        "django_settings_module": "{}.{}.settings".format(local_prefix, prjname),
        "site_domain": site_domain,
        "server_domain": server_domain,
        "server_url": server_url,
        "dev_packages": ' '.join([
            a.nickname for a in KNOWN_REPOS if a.nickname in dev_repos]),
        "pip_packages": ' '.join(pip_packages),
        "db_name": os.environ.get('GITLAB_CI_DB_NAME', None) or prjname,
        "python_path": sites_base,
        "usergroup": usergroup,
        "provider": "lino" if not ini_file or (ini_file and not use_django_settings) else "django"
    })

    if os.path.exists(join(project_dir, 'settings.py')):
        what = "reinstall {appname} to {project_dir}"
        reinstall = True
    else:
        what = "create {appname} site in {project_dir}"
        reinstall = False
    what = what.format(**context)
    click.echo("Preparing to {}".format(what))

    upstream_port = None
    used_ports = set()
    cfgp = ConfigParser()
    if asgi_server:
        if reinstall:
            cfgp.read(join(project_dir, 'lino.ini'))
            if cfgp.has_section('getlino_asgi') and cfgp.has_option(
                    'getlino_asgi', 'upstream_port'):
                try:
                    upstream_port = cfgp.getint('getlino_asgi',
                                                'upstream_port')
                except ValueError as e:
                    if cfgp.get('getlino_asgi', 'upstream_port') != "":
                        raise e
            cfgp.clear()
        if upstream_port is None:
            for p in os.listdir(python_path_root):
                if p != prjname:
                    p = join(python_path_root, p)
                    if os.path.isdir(p):
                        lini = join(p, 'lino.ini')
                        if os.path.exists(lini):
                            cfgp.read(lini)
                            try:
                                up = cfgp.getint('getlino_asgi',
                                                 'upstream_port')
                                used_ports.add(up)
                            except (NoSectionError, NoOptionError) as e:
                                pass
                            except ValueError as e:
                                if cfgp.get('getlino_asgi',
                                            'upstream_port') != "":
                                    raise e
                            cfgp.clear()
            upstream_port = 8001
            while True:
                if upstream_port in used_ports:
                    upstream_port += 1
                else:
                    break

        context.update(upstream_port=upstream_port)

    if not batch:
        shared_env = click.prompt("Shared virtualenv", default=shared_env)
        # if asroot:
        #     server_url = click.prompt("Server URL ", default=server_url)
        #     admin_name = click.prompt("Administrator's full name", default=admin_name)
        #     admin_email = click.prompt("Administrator's full name", default=admin_email)

    create_db = False
    settings_file = join(project_dir, "settings.py")
    if not os.path.exists(settings_file):
        if DEFAULTSECTION.get('db_user') and not db_user:
            shared_user = True
            db_user = DEFAULTSECTION.get('db_user')
            db_password = DEFAULTSECTION.get('db_password')
        else:
            shared_user = False
            db_user = db_user or DEFAULTSECTION.get('db_user')
            db_password = db_password or DEFAULTSECTION.get('db_password') \
                or secrets.token_urlsafe(8)
            if not db_user:
                db_user = prjname
            create_db = db_engine.needs_user

        if create_db and not batch:
            click.echo(
                "User credentials (for {db_engine} on {db_host}:{db_port}):".
                format(**context))
            db_user = click.prompt("- user name", default=db_user)
            db_password = click.prompt("- user password", default=db_password)
            if not GITLAB_CI:
                db_port = click.prompt("- port", default=db_port)
            db_host = click.prompt("- host name", default=db_host)

        if not batch:
            secret_key = click.prompt("Site's secret key", default=secret_key)

        context.update({
            "db_host": db_host,
            "db_port": db_port,
            "db_user": db_user,
            "db_password": db_password,
            "secret_key": secret_key,
        })

    if not i.yes_or_no("Okay to {}?".format(what)):
        raise click.Abort()

    os.umask(0o002)

    os.makedirs(project_dir, exist_ok=True)
    i.jinja_write(join(project_dir, "manage.py"), **context)

    if not os.path.exists(settings_file):
        tplname = None
        if ini_file and use_django_settings:
            tplname = 'django_settings.py'
        i.jinja_write(settings_file, tplname, **context)

    # pull.sh script is now in the virtualenv's bin folder
    # i.jinja_write(join(project_dir, "pull.sh"), **context)
    # if ifroot():
    #     log_link = join(project_dir, 'log')
    #     if not os.path.exists(log_link):
    #         logdir = join(DEFAULTSECTION.get("log_base"), prjname)
    #         os.makedirs(logdir, exist_ok=True)
    #         with i.override_batch(True):
    #             i.check_permissions(logdir)
    #             os.symlink(logdir, log_link)
    #             i.write_logrotate_conf('lino-{}.conf'.format(prjname),
    #                                    join(logdir, "lino.log"))

    bb = DEFAULTSECTION.get("backups_base")
    if bb and ifroot():
        if i.jinja_write(join(project_dir, "make_snapshot.sh"), **context):
            i.make_file_executable(join(project_dir, "make_snapshot.sh"))
            backups_base_dir = join(bb, prjname)
            os.makedirs(backups_base_dir, exist_ok=True)
            with i.override_batch(True):
                i.check_permissions(backups_base_dir)

            fn = 'make_snapshot_{prjname}'.format(**context)
            i.write_cron_job(fn,
                             MAKE_SNAPSHOT_CRON_SH.format(**context),
                             freq="daily")

    if web_server:
        if asgi_server:
            i.jinja_write(join(project_dir, "asgi.py"), **context)
        else:
            i.jinja_write(join(project_dir, "wsgi.py"), **context)
            pth = join(project_dir, web_server.name)
            os.makedirs(pth, exist_ok=True)
            if web_server.name == "nginx":
                i.jinja_write(join(pth, "uwsgi.ini"), **context)
                i.jinja_write(join(pth, "uwsgi_params"), **context)
        if hasattr(web_server, 'python_packages'):
            for p in web_server.python_packages.split():
                pip_packages.add(p)

    os.makedirs(join(project_dir, 'media'), exist_ok=True)

    i.check_virtualenv(envdir, context)

    if shared_env:
        pth = join(project_dir, DEFAULTSECTION.get('env_link'))
        if not os.path.exists(pth):
            os.symlink(envdir, pth)
        static_root = join(shared_env, 'static_root')
        if not os.path.exists(static_root):
            os.makedirs(static_root, exist_ok=True)

    if dev_repos:
        change_directory_into_dev_repos()
        click.echo("dev_repos is {} --> {}".format(dev_repos,
                                                   dev_repos.split()))
        repos = []
        for nickname in dev_repos.split():
            lib = REPOS_DICT.get(nickname, None)
            if lib is None:
                raise click.ClickException(
                    "Invalid repository nickname {} in --dev-repos".format(
                        nickname))
            repos.append(lib)
        click.echo("Installing {} repositories...".format(len(repos)))
        for lib in repos:
            i.clone_repo(lib)
        for lib in repos:
            i.install_repo(lib, envdir)

    if len(pip_packages):
        click.echo("Installing {} Python packages...".format(
            len(pip_packages)))
        i.run_in_env(
            envdir,
            "pip install -q --upgrade {}".format(' '.join(pip_packages)))

    if web_server:
        filename = "{}.conf".format(prjname)
        conf_root = join("/etc/", web_server.service)
        avpth = join(conf_root, 'sites-available', filename)
        enpth = join(conf_root, 'sites-enabled', filename)

        if asgi_server:

            if not os.path.exists('/run/daphne'):
                from pwd import getpwnam
                os.makedirs('/run/daphne')
                os.chown('/run/daphne',
                         getpwnam('root').pw_uid,
                         getpwnam('www-data').pw_gid)

                # below code re-creates the /run/daphne directory on server reboot
                i.write_file('/usr/lib/tmpfiles.d/daphne.conf',
                             'd /run/daphne 0755 root www-data\n')

            conf_tpl = 'nginx_asgi.conf'

            site_ini = join(project_dir, 'lino.ini')
            if not os.path.exists(site_ini):
                txt = JINJA_ENV.get_template('lino.ini').render()
                i.write_file(site_ini, txt)
                i.check_permissions(site_ini)
            channels = f"linod_{prjname}"
            cfgp.read(site_ini)
            if not cfgp.has_section('getlino_asgi'):
                cfgp.add_section('getlino_asgi')
            if cfgp.has_option('getlino_asgi', 'channels'):
                channels = cfgp.get('getlino_asgi', 'channels')
            else:
                cfgp.set('getlino_asgi', 'channels', channels)
            cfgp.set('getlino_asgi', 'upstream_port', str(upstream_port))
            if ini_file:
                cfgp.read_file(ini_file)
            with open(site_ini, 'r+') as f:
                cfgp.update_file(f)
            cfgp.clear()

            context.update(num_procs=os.cpu_count(),
                           channels=channels,
                           envdir=envdir)

            tpl = JINJA_ENV.get_template('daphne_supervisor.conf')
            conf = tpl.render(**context)
            i.write_supervisor_conf(f"{prjname}-asgi.conf", conf)
        else:
            conf_tpl = web_server.name + ".conf"
            # shutil.copyfile(join(project_dir, 'nginx', filename), avpth)
            if web_server.name == "nginx":
                i.write_supervisor_conf(
                    '{}-uwsgi.conf'.format(prjname),
                    UWSGI_SUPERVISOR_CONF.format(**context))

        if i.jinja_write(avpth, conf_tpl, **context):
            with i.override_batch(True):
                if i.check_overwrite(enpth):
                    os.symlink(avpth, enpth)
        i.must_restart(web_server.service)

    if create_db and db_engine.needs_user and not GITLAB_CI:
        if not shared_user:
            db_engine.setup_user(i, context)
        db_engine.setup_database(i, prjname, db_user, db_host)

    os.chdir(project_dir)
    i.run_in_env(envdir, "python manage.py install --noinput")

    if linod and not asgi_server and ifroot():
        fn = join(project_dir, 'linod.sh')
        i.write_file(fn, LINOD_SH.format(**context))
        i.check_permissions(fn, executable=True, writeable=False)
        i.write_supervisor_conf('linod_{}.conf'.format(prjname),
                                LINOD_SUPERVISOR_CONF.format(**context))
        i.must_restart('supervisor')

    if not reinstall:
        # i.run_in_env(envdir, "python manage.py migrate --noinput")  # Not necessary, prep does it from initdb.
        i.run_in_env(envdir, "python manage.py prep --noinput")
    db_engine.after_prep(i, context)
    if ifroot():
        i.run_in_env(envdir, "python manage.py buildcache -b")
        i.run_in_env(envdir, "python manage.py makehelp")

    i.run_apt_install()
    i.restart_services()

    if web_server:
        # I imagine that we need to actually restart nginx
        # before running certbot-auto because otherwise certbot would add
        # its entries to the default because it does does not yet see the
        # new site.

        if https:
            certbot_cmd = which_certbot()
            if certbot_cmd is None:
                raise click.ClickException("Oops, certbot is not installed.")
            i.runcmd("{} --{} -d {}".format(certbot_cmd, web_server.name,
                                            server_domain))
            i.must_restart(web_server.service)

    if ini_file:
        ini_file.close()
    click.echo("The new site {} has been created.".format(prjname))


params = [
    click.Argument(['appname'],
                   metavar="APPNAME",
                   type=click.Choice(APPNAMES),
                   required=False),
    click.Argument(['prjname'], required=False),
    click.Option(['--ini-file'],  # ini_file
                 type=click.File('r+'), required=False,
                 help="Read options from an existing lino.ini file."),
    click.Option(['--batch/--no-batch'], default=False, help=BATCH_HELP),
] + STARTSITE_OPTIONS
startsite = click.pass_context(startsite)
startsite = click.Command('startsite',
                          callback=startsite,
                          params=params,
                          help=startsite.__doc__)
